import json


class JSONSerializer(object):
    def _json(self, obj):
        return json.dumps(obj)

    def json(self, **kwargs):
        result = self.__call__(**kwargs)
        return self._json(result)


class DataFrameSerializer(JSONSerializer):
    def _json(self, data_frame):
        return data_frame.to_json()


class NumpySerializer(JSONSerializer):
    def _json(self, array):
        return json.dumps({"data": array.tolist()})



from copy import deepcopy
from functools import partial, wraps
from itertools import repeat
from inspect import Signature, signature

from dash.dependencies import Input, Output, State
from collections import deque, defaultdict


class Callback(object):
    def __init__(
        self,
        func,
        pre_func_hooks=None,
        post_func_hooks=None,
        inputs=None,
        outputs=None,
        states=None,
        to_api=True,
        tee=False,
    ):
        self.dash_inputs = self._to_dash(inputs, Input) or []
        self.dash_outputs = self._to_dash(outputs, Output) or []
        self.dash_states = self._to_dash(states, State) or []

        self._pre_function_hooks = pre_func_hooks or []
        self._post_function_hooks = post_func_hooks or []
        self.to_api = to_api
        self.current_inputs = []
        self.func = func
        self.is_provider = defaultdict(bool)
        self.tee = tee

    def unpack(self, f, args):
        if isinstance(args, tuple):
            return f(*args)
        return f(args)

    def pre_function_hook(self, inputs):
        for f in self._pre_function_hooks:
            inputs = self.unpack(partial(f, self), inputs)

        self.current_inputs = inputs
        return inputs

    def post_function_hook(self, outputs):
        for f in self._post_function_hooks:
            outputs = self.unpack(partial(f, self), outputs)

        return outputs

    def run(self, *inputs):
        if self._pre_function_hooks:
            function_inputs = self.pre_function_hook(inputs)
        else:
            function_inputs = inputs

        function_output = self.unpack(self.func, function_inputs)
        if self._post_function_hooks:
            return self.unpack(self.post_function_hook, function_output)

        if self.tee:
            return tuple(repeat(function_output, len(self.dash_outputs)))

        return function_output

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def register(self, outputs, inputs, states):
        self.set_inputs(inputs)
        self.set_outputs(outputs)
        self.set_states(states)

    def to(self, component, copy=False, **kwargs):
        if copy:
            callback = deepcopy(self)
        else:
            callback = self

        component.register_callback(callback, **kwargs)
        return callback

    def _from_dash(self, obj, **kwargs):
        if isinstance(obj, list):
            return [
                (
                    o.component_id.format(**kwargs),
                    o.component_property.format(**kwargs),
                )
                for o in obj
            ]
        return (
            obj.component_id.format(**kwargs),
            obj.component_property.format(**kwargs),
        )

    def _to_dash(self, obj, dash_type):
        if obj is None:
            return []

        if isinstance(obj, list):
            return [dash_type(id, prop) for id, prop in obj]

        return dash_type(obj[0], obj[1])

    def _format(self, **kwargs):
        self.dash_inputs = self._to_dash(
            self._from_dash(self.dash_inputs, **kwargs), Input
        )
        self.dash_outputs = self._to_dash(
            self._from_dash(self.dash_outputs, **kwargs), Output
        )
        self.dash_states = self._to_dash(
            self._from_dash(self.dash_states, **kwargs), State
        )

    def prepend_input(self, input):
        if isinstance(self.dash_inputs, Input):
            self.dash_inputs = [Input(input[0], input[1]), self.dash_inputs]
        else:
            self.dash_inputs.insert(0, Input(input[0], input[1]))

    def add_inputs(self, inputs):
        if inputs is not None:
            self.dash_inputs += [Input(id, prop) for id, prop in inputs]
        return self

    def set_inputs(self, inputs):
        if inputs is not None:
            self.dash_inputs = self._to_dash(inputs, Input)
        return self

    def add_outputs(self, outputs):
        if outputs is not None:
            new_outputs = [Output(id, prop) for id, prop in outputs]
            if isinstance(self.dash_outputs, list):
                self.dash_outputs += new_outputs
            else:
                self.dash_outputs = [self.dash_outputs] + new_outputs
        return self

    def set_outputs(self, outputs):
        if outputs is not None:
            self.dash_outputs = self._to_dash(outputs, Output)
        return self

    def add_states(self, states):
        if states is not None:
            self.dash_states += [State(id, prop) for id, prop in states]
        return self

    def set_states(self, states):
        if states is not None:
            self.dash_states = self._to_dash(states, State)
        return self

    def add_figure_output(self, figure):
        figure.register_callback(self)
        return self

    def add_storage_output(self, storage, key=None, copy=False):
        if copy:
            callback = deepcopy(self)
        else:
            callback = self

        storage.add_provider(callback, key)
        return callback

    def use_storage_inputs(self, storage, *keys, copy=False, unpack_args=False):
        if copy:
            callback = deepcopy(self)
        else:
            callback = self

        keys = keys or storage.keys

        self.dash_inputs = [Input(storage.id, f"trigger-{key}") for key in keys]
        self.dependencies = keys or []
        storage.add_client(callback, unpack_args)

        return callback

    def register_api(self, app):
        app.callbacks.add(self)

    @property
    def description(self):
        return {
            "name": self.name,
            "parameters": self.parameters,
            "docstring": self.func.__doc__,
            "return_type": self.return_type,
        }

    @property
    def name(self):
        return self.func.__name__

    @property
    def parameters(self):
        sig = signature(self.func)
        params = []
        for name, param in sig.parameters.items():
            param_doc = {"name": name}
            if param.annotation is not param.empty:
                param_doc["type"] = str(param.annotation)
            if param.default is not param.empty:
                param_doc["default"] = param.default

            params.append(param_doc)

        return params

    @property
    def return_type(self):
        t = signature(self.func).return_annotation
        if t != Signature.empty:
            return t
        return


def as_callback(
    *mixins,
    pre_func_hooks=None,
    post_func_hooks=None,
    inputs=None,
    outputs=None,
    states=None,
    to_api=True,
    tee=False,
):
    def wrapper(func, *args, **kwargs):
        @wraps(func)
        def inner(*args, **kwargs):
            return func(*args, **kwargs)

        class NewCallback(Callback, *mixins):
            pass

        return NewCallback(
            inner,
            pre_func_hooks=pre_func_hooks,
            post_func_hooks=post_func_hooks,
            inputs=inputs,
            outputs=outputs,
            states=states,
            to_api=to_api,
            tee=tee,
        )

    return wrapper



import os

import plotly.express as px
import pandas as pd

from .figure import Figure

class PandasScatterMap(Figure):
    def __init__(self, id, key, mapbox_token, dataset_config=None, **kwargs):
        dataset_config = dataset_config or None
        super().__init__(id, keys=[key], **kwargs)
        self.dataset_config = dataset_config
        self.mapbox_token = mapbox_token

    def _process_data(self, data):
        return data[self.keys[0]]

    def display(self, df):
        px.set_mapbox_access_token(self.mapbox_token)
        if df.empty:
            dummy = pd.DataFrame(dict(lat=[0], lon=[0]))
            return px.scatter_mapbox(dummy, "lat", "lon", mapbox_style="white-bg").to_dict()
        fig = px.scatter_mapbox(df, **self.dataset_config)
        fig_dict = fig.to_dict()
        return fig_dict



# isort: skip

from .core import (
    Callback,
    Component,
    KWSDashApp,
    KWSSuperApp,
    as_callback,
    as_component,
)

from .components import ComponentGroup, Figure, PandasFigure, PandasTable



import codecs
import os
import pickle
from collections import deque
from tempfile import TemporaryDirectory, TemporaryFile
from uuid import uuid4

from dash.dependencies import Input, Output
from dash_html_components import Div
from dash_infra.core import Callback, Component, as_callback
from flask import request
from flask import current_app as app
from datetime import datetime


class Storage(Component):
    """
    A storage component modifies a callback so
    it writes into a single serialized dictionary.
    This dictionary (state) has one key per registered
    callback, storing the output of the callback.
    """

    def __init__(self, id, use_inner_store=False):
        super().__init__(id)
        self.use_inner_store = use_inner_store
        self.keys = []
        self.clients = []
        self.providers = []

    def layout(self):
        return Div(id=self.id)

    def _serializer(self, *outputs):
        raise NotImplementedError

    def _deserializer(self, client_id, trigger_keys, states):
        raise NotImplementedError

    def _trigger(self):
        return str(uuid4())

    def serializer(self, client_id, callback_id, *outputs):
        if not outputs:
            outputs = ""

        return self._serializer(client_id, callback_id, *outputs)

    def add_provider(self, callback, copy=False):
        callback.is_provider[self.id] = True

        def pre_hook(state, client_id, *inputs):
            state.client_id = client_id
            return inputs

        def post_hook(state, callback_output):
            if callback_output is not None:
                return self.serializer(state.client_id, callback.name, callback_output)
            else:
                return {}

        callback._pre_function_hooks.insert(0, pre_hook)
        callback._post_function_hooks.append(post_hook)
        self.providers.append(callback)

        return callback

    def _add_store_prehook(self, callback):
        if not callback.is_provider[self.id]:

            def pre_hook(s, client_id, *triggers_states_inputs):
                num_triggers = len(callback.dependencies)
                trigger_keys = callback.dependencies

                if self.use_inner_store:
                    states = triggers_states_inputs[num_triggers : 2 * num_triggers]
                    other_inputs = triggers_states_inputs[2 * num_triggers :]
                else:
                    states = []
                    other_inputs = triggers_states_inputs[num_triggers:]

                result = self._deserializer(client_id, trigger_keys, states)

                if callback.unpack_args:
                    return tuple([r for r in result.values()] + list(other_inputs))
                else:
                    return tuple([result] + list(other_inputs))

        else:

            def pre_hook(s, *triggers_states_inputs):
                num_triggers = len(callback.dependencies)
                trigger_keys = callback.dependencies

                if self.use_inner_store:
                    states = triggers_states_inputs[num_triggers : 2 * num_triggers]
                    other_inputs = triggers_states_inputs[2 * num_triggers :]
                else:
                    states = []
                    other_inputs = triggers_states_inputs[num_triggers:]

                result = self._deserializer(callback.client_id, trigger_keys, states)
                if callback.unpack_args:
                    return tuple([r for r in result.values()] + list(other_inputs))
                else:
                    return tuple([result] + list(other_inputs))

        callback._pre_function_hooks.append(pre_hook)

    def add_client(self, callback, unpack_args=False):
        callback.unpack_args = unpack_args
        self.clients.append(callback)

    def _before_registry(self, app):
        super()._before_registry(app)
        for provider in self.providers:
            provider.prepend_input(("main-div", "title"))
            provider.set_outputs((self.id, f"trigger-{provider.name}"))
            self.keys.append(provider.name)

            if self.use_inner_store:
                provider.add_outputs([(self.id, f"value-{provider.name}")])

        for client in self.clients:
            if not client.dependencies:
                client.dependencies = self.keys

            old_inputs = client.dash_inputs
            client.set_inputs(
                [("main-div", "title")]
                + [(self.id, f"trigger-{key}") for key in client.dependencies]
            )
            client.dash_inputs = client.dash_inputs + [
                i for i in old_inputs if i not in client.dash_inputs
            ]

            if self.use_inner_store:
                client.set_states(
                    [
                        (self.id, f"value-{key}")
                        for key in client.dependencies
                        if key != client.name
                    ]
                )

            self._add_store_prehook(client)

    def register_callback(self, callback, **kwargs):
        return self.add_provider(callback, **kwargs)


class PickleStorage(Storage):
    def __init__(self, id):
        super().__init__(id, use_inner_store=True)

    def _serializer(self, client_id, callback_id, *outputs):
        if len(outputs) == 1:
            outputs = outputs[0]

        serialized = codecs.encode(pickle.dumps(outputs), "base64").decode()
        return self._trigger(), serialized

    def _deserializer(self, client_id, trigger_names, states):
        obj = {}
        for name, state in zip(trigger_names, states):
            obj[name] = pickle.loads(codecs.decode(state.encode(), "base64"))
        return obj


class FileStorage(Storage):
    import tempfile

    def __init__(self, id, maxsize=20):
        super().__init__(id)
        self.stores = deque(maxlen=maxsize)
        self.maxsize = maxsize
        self._directory = TemporaryDirectory()
        self.directory = self._directory.name

    def get_store(self, client_id):
        if client_id not in self.stores:
            if len(self.stores) == self.maxsize:
                last_id = self.stores.popleft()
                os.remove(os.path.join(self.directory, last_id))
            self.stores.append(client_id)

        store_path = os.path.join(self.directory, client_id)
        return store_path

    def _serializer(self, client_id, callback_id, *outputs):
        path = self.get_store(client_id)
        with open(path, "w+b") as fp:
            if len(outputs) == 1:
                outputs = outputs[0]

            if os.path.getsize(path) > 0:
                state = pickle.load(fp)
                state[callback_id] = outputs
            else:
                state = {callback_id: outputs}

            fp.seek(0)
            pickle.dump(state, fp)

        return self._trigger()

    def _deserializer(self, client_id, trigger_names, _):
        path = self.get_store(client_id)
        with open(path, "rb") as fp:
            obj = pickle.load(fp)
            return {name: obj[name] for name in trigger_names}


class ContextStorage(Storage):
    def __init__(self, id, maxsize=20):
        super().__init__(id, use_inner_store=False)
        self.stores = deque(maxlen=maxsize)
        self.maxsize = maxsize

    def _before_registry(self, app):
        super()._before_registry(app)
        app.server.store = {}

    def get_store(self, client_id):
        stores = app.store
        if client_id not in self.stores:
            stores[client_id] = {}
            if len(self.stores) == self.maxsize:
                last_id = self.stores.popleft()
                del stores[last_id]
            self.stores.append(client_id)

        return stores[client_id]

    def _serializer(self, client_id, callback_id, *outputs):
        store = self.get_store(client_id)
        if len(outputs) == 1:
            outputs = outputs[0]
        store[callback_id] = outputs
        return self._trigger()

    def _deserializer(self, client_id, trigger_names, _):
        store = self.get_store(client_id)
        return {name: store.get(name) for name in trigger_names}



def CustomElement(elem):
    def wrapper(*args, classes=None, **kwargs):
        if classes:
            class_name = kwargs.pop("className", "") + " " + " ".join(classes)
            return elem(*args, **kwargs, className=class_name)

        else:
            return elem(*args, **kwargs)

    return wrapper



import json
from collections import defaultdict
from typing import Dict

import click

from flask import Blueprint, Flask, jsonify, render_template, request
from slugify import slugify
from werkzeug.middleware.dispatcher import DispatcherMiddleware


class KWSSuperApp:
    def __init__(
        self,
        dash_router: Dict[str, "KWSDashApp"],
        *args,
        name="Default Application",
        **kwargs,
    ):
        self.app = Flask(__name__, *args, **kwargs)
        self.name = name
        self.api = Blueprint("API", __name__, url_prefix="/api/")
        self.router = dash_router
        self.api_endpoints = defaultdict(dict)

        self.register_callbacks_as_api()

    def register_slug(self, name):
        slug = slugify(name)
        return f"/{slug}/"

    def register_callbacks_as_api(self):
        for app in self.router.values():
            for callback in app.callbacks:
                self.register_api(callback)

    def register_api(self, callback):
        if not hasattr(callback, "_json") or not callback.to_api:
            return

        slug = slugify(callback.func.__name__)
        route = self.register_slug(callback.func.__name__)

        def wrapped_function():
            payload = request.form or json.loads(request.json)
            result = callback.json(**payload)
            return jsonify(result), 200

        def wrapped_doc():
            doc = {"docstring": callback.func.__doc__}

            return jsonify(doc), 200

        wrapped_function.__name__ = f"{slug}_worker"
        wrapped_doc.__name__ = f"{slug}_doc"
        self.api.route(route, methods=("POST",))(wrapped_function)
        self.api.route(route, methods=("GET",))(wrapped_doc)
        self.api_endpoints[slugify(callback.name)] = callback.description

        return callback

    def create_app(self):
        import math

        def slice_array(array, length):
            return [
                array[i * length : (i * length) + length]
                for i in range(math.ceil(len(array) / length))
            ]

        self.app.register_blueprint(self.api)

        reshaped_endpoints = []
        for row in slice_array(list(self.api_endpoints.keys()), 3):
            reshaped_endpoints.append({n: self.api_endpoints[n] for n in row})

        @self.app.route("/endpoints/")
        def endpoints():
            return render_template(
                "endpoints.html",
                endpoints=reshaped_endpoints,
            )

        @self.app.route("/")
        def index():

            return render_template(
                "index.html",
                name=self.name,
                router=self.router,
                endpoints=reshaped_endpoints,
            )

        for route, dash_app in self.router.items():
            dash_app.instanciate(
                server=self.app, routes_pathname_prefix=f"/{route}/",
            )
        return self.app



    @property
    def cli(self):
        @click.command()
        @click.option(
            "--name", "-n", required=True, type=click.Choice(self.router.keys())
        )
        @click.option("--debug", "-d", is_flag=True)
        def run_app(name, debug):
            app = self.router[name].instanciate()
            app.run_server(debug=debug)

        @click.command()
        @click.option("--debug", "-d", is_flag=True)
        def run(debug):
            app = self.create_app()
            app.run(debug=debug)

        return click.Group("run", {"run_app": run_app, "run": run})



import os
from uuid import uuid4
from collections import deque
from tempfile import TemporaryDirectory
from zipfile import ZipFile

from flask import send_from_directory

from dash_infra.core import Component
from dash_html_components import A


class DownloadZip(Component):
    def __init__(self, id, maxsize=10):
        self._upload_folder = TemporaryDirectory()
        self.upload_folder = self._upload_folder.name
        self.maxsize = maxsize
        self.tmpfiles = deque(maxlen=maxsize)
        super().__init__(id)

    def layout(self):
        return A(id=self.id, children=f"Download {self.id}", href="", className="btn")

    def _before_registry(self, app):
        @app.server.route(f"/download/{self.id}/<filename>")
        def send(filename):
            return send_from_directory(self.upload_folder, filename)

    def register_callback(self, callback, copy=False):
        def post_hook(state, callback_output):
            if len(self.tmpfiles) == self.maxsize:
                last_id = self.tmpfiles.popleft()
                os.remove(os.path.join(self.upload_folder, last_id))

            file_id = str(uuid4()) + ".zip"
            self.tmpfiles.append(file_id)

            with ZipFile(os.path.join(self.upload_folder, file_id), "w") as zipf:
                for fname, fcontent in callback_output.items():
                    with zipf.open(fname, "w") as fp:
                        fp.write(fcontent.encode("utf-8"))

            return f"/download/{self.id}/{file_id}"

        callback._post_function_hooks.append(post_hook)
        callback.set_outputs((self.id, "href"))
        self.callbacks.add(callback)



import pandas as pd
from dash_infra.core import Component, as_callback
from dash_table import DataTable

from .store_user import StoreUser


class PandasTable(StoreUser):
    """
    A Figure component uses a Storage Component as a temporary
    output for all registered callback inputs.
    Each output is stored as a key in the state of the output
    and can be accessed as a dataset.
    """

    def __init__(self, id, key, store=None, **kwargs):
        super().__init__(
            id,
            display_component=DataTable,
            display_properties=["columns", "data"],
            store=store,
            keys=[key],
            **kwargs
        )

    def _process_data(self, data):
        return data[self.keys[0]]

    def display(self, df):
        columns = [{"name": i, "id": i} for i in df.columns]
        data = df.to_dict("records")
        return columns, data
